unit LZSS;

interface

uses Windows, Classes, SysUtils;

Const
  TLZ_LITERAL       = 1;
  TLZ_MATCH         = 0;
  TLZ_MINMATCHLEN   = 2;
  TLZ_WINBITS       = 10;
  TLZ_MATCHBITS     = 4;
  TLZ_WINLEN        = 1023;
  TLZ_MATCHLEN      = 17;

type
  TLZ = Class
  public
    function Compress(const Source; var Dest; Size: ULONG): ULONG;
    function Decompress(const Source; var Dest; Size: PULONG): ULONG;
  private

    DataStreamPos: ULONG;
    CompressedStreamPos: ULONG;
    Data:array of UCHAR;
    CompressedData:array of UCHAR;

    DataSize: ULONG;
    CompressedSize: ULONG;
    CompressedLong: ULONG;
    CompressedBitsUsed: Integer;

    //Functions

    function CompressLoop(): ULONG;
    procedure FindMatches(Offset, Len: PUINT);
    procedure CompressedStreamReadBits(Value: PUINT; NumBits: UINT);
    procedure CompressedStreamWriteBits(Value : UINT; NumBits: UINT);
    procedure CompressedStreamWriteBitsFlush();
end;




implementation


{ TLZ }

function TLZ.Compress(const Source; var Dest; Size: ULONG): ULONG;
begin
  //Set up initial values
  DataStreamPos:=        0;
  CompressedStreamPos:=  0;
  CompressedLong:=       0;
  CompressedBitsUsed:=   0;
  CompressedData:=       @Dest;
  Data:=                 @Source;
  DataSize :=            Size;
  CompressedSize:=       0;
  
  CompressLoop;

  CompressedStreamWriteBitsFlush;
  Result := CompressedSize;
end;

procedure TLZ.CompressedStreamReadBits(Value: PUINT; NumBits: UINT);
var
  Temp: ULONG;
begin
  CompressedLong:= CompressedLong and $FFFF;
  while NumBits > 0 do
  begin
    dec(NumBits);
    if CompressedBitsUsed = 0 then
    begin
      Temp:= CompressedData[CompressedStreamPos];
      inc(CompressedStreamPos);
      CompressedLong:= CompressedLong or (Temp shl 8);
      Temp:= CompressedData[CompressedStreamPos];
      inc(CompressedStreamPos);
      CompressedLong:= CompressedLong or Temp;
      CompressedBitsUsed:= 16;
    end;
    CompressedLong:= CompressedLong shl 1;
    dec(CompressedBitsUsed);
  end;
  Value^:= CompressedLong shr 16;
end;

procedure TLZ.CompressedStreamWriteBits(Value, NumBits: UINT);
begin
  while NumBits > 0 do
  begin
    dec(NumBits);
    CompressedLong := CompressedLong shl 1;
    CompressedLong:= CompressedLong or ((Value shr NumBits) and 1);
    inc(CompressedBitsUsed);

    if(CompressedBitsUsed = 32) then
    begin
      CompressedData[CompressedStreamPos]:= CompressedLong shr 24;
      CompressedData[CompressedStreamPos+1]:= CompressedLong shr 16;
      inc(CompressedStreamPos,2);
      inc(CompressedSize,2);
      dec(CompressedBitsUsed,16);
    end;
  end;  //end while
end;

procedure TLZ.CompressedStreamWriteBitsFlush;
begin
  CompressedLong:= CompressedLong shl (32 - CompressedBitsUsed);
  case CompressedBitsUsed of
  1..8:
    begin
      CompressedData[CompressedStreamPos]:= CompressedLong shr 24;
      inc(CompressedSize);
      inc(CompressedStreamPos);
    end;
  9..16:
    begin
      CompressedData[CompressedStreamPos]:= CompressedLong shr 24;
      CompressedData[CompressedStreamPos+1]:= CompressedLong shr 16;
      inc(CompressedStreamPos, 2);
      inc(CompressedSize, 2);
    end;
  17..24:
    begin
      CompressedData[CompressedStreamPos]:= CompressedLong shr 24;
      CompressedData[CompressedStreamPos+1]:= CompressedLong shr 16;
      CompressedData[CompressedStreamPos+2]:= CompressedLong shr 8;
      inc(CompressedStreamPos, 3);
      inc(CompressedSize, 3);
    end;
  25..32:
    begin
      CompressedData[CompressedStreamPos]:= CompressedLong shr 24;
      CompressedData[CompressedStreamPos+1]:= CompressedLong shr 16;
      CompressedData[CompressedStreamPos+2]:= CompressedLong shr 8;
      CompressedData[CompressedStreamPos+3]:= CompressedLong;
      inc(CompressedStreamPos, 4);
      inc(CompressedSize, 4);
    end;
  end;
      CompressedBitsUsed:= 0;

end;

function TLZ.CompressLoop: ULONG;
const
  Max = 128;
var // Optimization variables
  f: array[0..Max] of UINT;
  p: array[0..Max] of UINT;
  k, m: integer;
var
  Offset, MatchLength: UINT;
begin
  while DataStreamPos < DataSize do
  begin
    FindMatches(@Offset, @MatchLength);
    if MatchLength < TLZ_MINMATCHLEN then
    begin
      CompressedStreamWriteBits(TLZ_LITERAL, 1);
      CompressedStreamWriteBits(Data[DataStreamPos], 8);
      inc(DataStreamPos);
    end
    else
    begin
      f[0]:= DataStreamPos;
      f[1]:= DataStreamPos + 1;
      m:= 2;
      while true do
      begin
        p[m-2]:= Offset;
        f[m]:= f[m-2] + MatchLength;
        if (f[m] < f[m-1]+1) then break;
        inc(m);
        if Succ(m) = Max then
          break;
        DataStreamPos:= f[m-2];
        FindMatches(@Offset, @MatchLength);
      end;
      if((m and 1) = 0) then
      begin
        CompressedStreamWriteBits(TLZ_LITERAL, 1);
        CompressedStreamWriteBits(Data[f[0]], 8);
      end;
      k:= (m+1) and 1;
      while K < m-1 do
      begin
        CompressedStreamWriteBits(TLZ_MATCH, 1);
        CompressedStreamWriteBits(p[k], TLZ_WINBITS);
        CompressedStreamWriteBits(f[k+2] - f[k] - TLZ_MINMATCHLEN, TLZ_MATCHBITS);
        DataStreamPos:= f[k+2];
        inc(k,2);
      end;
    end;
  end;  // end while
  CompressedStreamWriteBits(TLZ_MATCH, 1);
  CompressedStreamWriteBits(0,TLZ_WINBITS); // Flag  
end;  // Compress Loop

procedure TLZ.FindMatches(Offset, Len: PUINT);
var
  WPos: ULONG;
  TempLen: ULONG;
  MaxWPos: ULONG;
begin
  Len^:= TLZ_MINMATCHLEN - 1;
  if DataStreamPos < TLZ_WINLEN then
    WPos := 0
  else
    WPos:= DataStreamPos - TLZ_WINLEN;
  MaxWPos:= DataStreamPos;
  while (WPos < MaxWPos) do
    begin
      TempLen:= 0;
      while((Data[WPos + TempLen] = Data[DataStreamPos + TempLen]) and ((DataStreamPos + TempLen) < DataSize) and (TempLen < TLZ_MATCHLEN)) do
      begin
        inc(TempLen);
      end;  // end while
      if (TempLen > Len^) and ((WPos and TLZ_WINLEN) <> 0) then
      begin
        Len^:= TempLen;
        Offset^:= WPos;
      end;
      inc(WPos);
    end;
end;

function TLZ.Decompress(const Source; var Dest; Size: PULONG): ULONG;
var
  Temp: UINT;
  Offset, Len: UINT;
  Buffer: array[0..TLZ_WINLEN] of Byte;
begin
  DataStreamPos:= 0;
  CompressedStreamPos:= 0;
  CompressedLong:= 0;
  CompressedBitsUsed:= 0;
  CompressedData:= @Source;
  Data:= @Dest;

  while true do
  begin
    CompressedStreamReadBits(@Temp,1);

    if Temp = TLZ_LITERAL then
    begin
      CompressedStreamReadBits(@Temp, 8);
      Data[DataStreamPos]:= Temp;
      Buffer[DataStreamPos and TLZ_WINLEN]:= Temp;
      inc(DataStreamPos);
    end
    else
    begin
      CompressedStreamReadBits(@Offset,TLZ_WINBITS);
      if Offset = 0 then break;
      CompressedStreamReadBits(@Len, TLZ_MATCHBITS);
      Len:= Len + TLZ_MINMATCHLEN;
      while Len > 0 do
      begin
        dec(Len);
        Data[DataStreamPos]:= Buffer[Offset and TLZ_WINLEN];
        Buffer[DataStreamPos and TLZ_WINLEN]:= Data[DataStreamPos];
        inc(DataStreamPos);
        inc(Offset);
      end;
    end;
  end;
  Size^:= CompressedStreamPos;
  Result:= DataStreamPos;
end;

end.
